// Copyright (c) 2016 Citrix Systems, Inc. All Rights Reserved.
// You may only reproduce, distribute, perform, display, or prepare derivative works of this file pursuant to a valid license from Citrix.
// THIS SAMPLE CODE IS PROVIDED BY CITRIX "AS IS" AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.

#include <sstream>

#include "PostCertificateAndTokenRequest.h"
#include "Utilities.h"

using namespace CitrixAuthManagerCustomProtocolSDK;

const std::string CitrixCertificateHeader = "X-Citrix-AM-ClientCertificateDeclaration: ";
const std::string ContentTypeHeader       = "Content-Type: ";
const std::string ContentTypeHeaderValue  = "application/vnd.citrix.requesttoken+xml";
const std::string AcceptHeader            = "Accept:";
const std::string AcceptHeaderValue       = "application/vnd.citrix.requesttokenresponse+xml";
const std::string HeaderDelimiter         = "\r\n";

std::string BuildCertificateHeader(const std::vector<char>& certificateFileContent)
{
    std::string base64_encoded_certificate;
    DataToBase64(certificateFileContent, base64_encoded_certificate);
    return CitrixCertificateHeader + base64_encoded_certificate;
}

std::string BuildContentTypeHeader()
{
    return ContentTypeHeader + ContentTypeHeaderValue;
}

std::string BuildAcceptHeader()
{
    return AcceptHeader + AcceptHeaderValue;
}

// Create a string of new line delimited headers combining:
//   The content type header.
//   The user identifying certificate data header with the certificate data as a base 64 encoded string.
std::string BuildAdditionalHeaders(const std::vector<char>& certificateFileContent)
{
    return BuildCertificateHeader(certificateFileContent) + '\n' + BuildContentTypeHeader() + '\n' + BuildAcceptHeader();
}

PostCertificateAndTokenRequest::PostCertificateAndTokenRequest() :
    m_curlEasyHandle(NULL),
    m_headers(NULL)
{
    // Initialize curl.
    curl_global_init(CURL_GLOBAL_DEFAULT);
}

PostCertificateAndTokenRequest::~PostCertificateAndTokenRequest()
{
    CloseHandles();

    // Cleanup curl.
    curl_global_cleanup();
}

eAuthResult PostCertificateAndTokenRequest::PostRequestAndCertificateToAuthExtension(const std::string& url,
                                                                                     const std::string& tokenRequest,
                                                                                     const std::vector<char>& certificateFileContent,
                                                                                     /* out */ std::vector<char>& outServerResponse)
{
    eAuthResult result = PostRequestAndCertificateToAuthExtensionInternal(url, tokenRequest, certificateFileContent, outServerResponse);

    CloseHandles();

    return result;
}

eAuthResult PostCertificateAndTokenRequest::PostRequestAndCertificateToAuthExtensionInternal(const std::string& startURL,
                                                                                             const std::string& tokenRequest,
                                                                                             const std::vector<char>& certificateFileContent,
                                                                                             /* out */ std::vector<char>& outServerResponse)
{
    TRACE_WITH_ARGS("Authentication URL to POST to is '%s'", startURL.c_str());

    // Initialize the output parameter.
    outServerResponse.clear();

    bool is_https = false;
    unsigned int port = 0;
    std::string host_name;
    std::string extra;
    if (!CrackUrl(startURL, is_https, port, host_name, extra))
    {
        TRACE("Failed to crack URL");

        return SystemError;
    }
    TRACE("URL cracked");

    // For this example we will assume that the plugin has a security policy that
    // requires HTTPS for secure transport for the conversation with the server.
    // Your plugin may have other security requirements and may want to signal
    // to AM and Receiver that there was a security problem.
    if (!is_https)
    {
        TRACE("The provided URL is not HTTPS");

        return DisallowedBySecurityPolicy;
    }

    // Initialize the curl handle.
    m_curlEasyHandle = curl_easy_init();
    if (m_curlEasyHandle == NULL)
    {
        TRACE("\"curl_easy_init\" failed");

        return SystemError;
    }
    TRACE("HTTP machinery initialized");

    // Set the method.
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_CUSTOMREQUEST, "POST");

    // Set the user agent.
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_USERAGENT, "A Citrix Auth Manager plugin example");

    // Set the URL and port.
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_URL, startURL.c_str());
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_PORT, port);

    // Set the timeout just in case.
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_TIMEOUT, 30);

    // HTTPS settings.
    if (is_https)
    {
        // Use SSL.
        curl_easy_setopt(m_curlEasyHandle, CURLOPT_USE_SSL, CURLUSESSL_ALL);

        // We don't want a pre-installed OS bundle of certificates.
        curl_easy_setopt(m_curlEasyHandle, CURLOPT_CAINFO, NULL);

        // Set the directory to be searched for root certificates.
        std::string certificates_path = GetIcaRoot() + "/keystore/cacerts";
        curl_easy_setopt(m_curlEasyHandle, CURLOPT_CAPATH, certificates_path.c_str());

        // Automatically fetch the certificates info.
        curl_easy_setopt(m_curlEasyHandle, CURLOPT_CERTINFO, 1L);

        // Disconnect if we can't validate server's certificate.
        curl_easy_setopt(m_curlEasyHandle, CURLOPT_SSL_VERIFYPEER, 1L);
    }

    // Build the headers with the client certificate header and content type.
    m_headers = curl_slist_append(m_headers, BuildCertificateHeader(certificateFileContent).c_str());
    m_headers = curl_slist_append(m_headers, BuildContentTypeHeader().c_str());
    m_headers = curl_slist_append(m_headers, BuildAcceptHeader().c_str());
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_HTTPHEADER, m_headers);

    // Set the POST data.
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_POSTFIELDS, tokenRequest.c_str());
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_POSTFIELDSIZE, tokenRequest.size());

    // Set the response callbacks.
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_HEADERDATA, this);
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_HEADERFUNCTION, writeHeaderHandle);
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_WRITEDATA, this);
    curl_easy_setopt(m_curlEasyHandle, CURLOPT_WRITEFUNCTION, writeResponseHandle);

    // Set any more curl options here. E.g. Proxy settings etc.

    TRACE("HTTP request prepared");

    // Perform the request.
    CURLcode result = curl_easy_perform(m_curlEasyHandle);
    if (result != CURLE_OK)
    {
        TRACE_WITH_ARGS("\"curl_easy_perform\" failed: %s", curl_easy_strerror(result));

        return ProtocolError;
    }
    TRACE("HTTP request performed");

    // Get the response code.
    unsigned long status = 0;
    curl_easy_getinfo(m_curlEasyHandle, CURLINFO_RESPONSE_CODE, &status);
    TRACE_WITH_ARGS("HTTP Status code is %lu", status);
    switch (status)
    {
    case 200:
        // OK.

        break;
    case 401:
        // Unauthorized.
        // This is returned by the StoreFront extension if the certificate could not be mapped
        // to a valid user account or some basic account checks failed.
        TRACE("Response is 401: unauthorized");

        return AccessDenied;
    case 404:
        // Not found.
        TRACE("Response is 404: not found");

        return NetworkError;
    case 403:
        // Forbidden.
        // This is returned by the StoreFront extension if the certificate could not be parsed
        // or there was no certificate sent in the request.
        TRACE("Response is 403: forbidden");

        return ProtocolError;
    default:
        // Anything else is assumed to be some form of server error.
        TRACE_WITH_ARGS("Response is %lu: server error", status);

        return AuthenticationServerError;
    }

    // Check the returned headers.
    if (m_responseHeaders.empty())
    {
        TRACE("No response headers: failed to get the content type");

        return ProtocolError;
    }

    TRACE("Got back a content type header");

    // Check that the returned content type is what is expected.
    bool isResponseValid = false;
    for (std::vector<std::string>::iterator it = m_responseHeaders.begin(); it != m_responseHeaders.end(); it++)
    {
        std::string header = *it;
        size_t header_end = header.find(HeaderDelimiter);
        header = header.substr(0, header_end);
        if (header.compare(0, ContentTypeHeader.length(), ContentTypeHeader) == 0)
        {
            header = header.substr(ContentTypeHeader.length(), header_end);
            if (header.compare(AcceptHeaderValue) == 0)
            {
                isResponseValid = true;

                break;
            }
        }
    }
    if (!isResponseValid)
    {
        TRACE("Wrong content type!");

        return ProtocolError;
    }
    TRACE("Content type header value is correct");

    // Copy the response data to the ouput buffer.
    outServerResponse.assign(m_responseData.begin(), m_responseData.end());
    if (outServerResponse.size() == 0)
    {
        TRACE("The response had an empty body");

        return ProtocolError;
    }

    TRACE_WITH_ARGS("Content is of length %zu", outServerResponse.size());

    // Success.
    // A response with a body and the correct content type was received.
    // Handles are closed outside this scope to allow easy early out on an error.
    return Success;
}

void PostCertificateAndTokenRequest::CloseHandles()
{
    if (m_headers != NULL)
    {
        // Free the headers.
        curl_slist_free_all(m_headers);
        m_headers = NULL;
    }

    if (m_curlEasyHandle != NULL)
    {
        // Cleanup curl.
        curl_easy_cleanup(m_curlEasyHandle);
        m_curlEasyHandle = NULL;
    }
}

size_t PostCertificateAndTokenRequest::writeHeaderHandleImpl(char* buffer, size_t size, size_t nmemb)
{
    // Get the header string.
    size_t header_size = size * nmemb;
    std::string header(buffer, header_size);

    // Append the header to the list of response headers.
    m_responseHeaders.push_back(header);

    return header_size;
}

size_t PostCertificateAndTokenRequest::writeHeaderHandle(char* buffer, size_t size, size_t nmemb, void* userp)
{
    return static_cast<PostCertificateAndTokenRequest*>(userp)->writeHeaderHandleImpl(buffer, size, nmemb);
}

size_t PostCertificateAndTokenRequest::writeResponseHandleImpl(char* buffer, size_t size, size_t nmemb)
{
    // Append the data to the whole response.
    size_t response_size = size * nmemb;
    m_responseData.append(buffer, response_size);
    return response_size;
}

size_t PostCertificateAndTokenRequest::writeResponseHandle(char* buffer, size_t size, size_t nmemb, void* userp)
{
    return static_cast<PostCertificateAndTokenRequest*>(userp)->writeResponseHandleImpl(buffer, size, nmemb);
}
